import java.io.*;
import java.util.*;

/**
 * Class for Image objects.
 * Supports file IO for Portable Graymaps (PGM) and 8-, 16-, or 24-bit Bitmaps (BMP)
 * 
 * @author Dane Barney
 * @version 2004.02.12
 * 
 * !!!! FOR USE IN PROJECT 2 OF CSE490GZ !!!!
 */
public class Image {
    public static final int BITMAP = 0;
    public static final int PGM = 1;

    public int height;
    public int width;
    public short bitDepth;
    public int type;
    public int filesize;
    public byte[] header;
    public int[][] pixels;
    public int[] colorTable; // stores the actual pixel colors as RGB values
    
    /**
     * Create an image object from a file
     */
    public Image(String filename) throws IOException {
        DataInputStream in = new DataInputStream(new FileInputStream(filename));
        
        getType(filename);
        
        switch(type) {
            case BITMAP:
                readBitmapHeader(in);
                pixels = new int[width][height];
                fillBitmapPixels(in);
                break;
            case PGM:
                BufferedReader rd = new BufferedReader(new FileReader(filename));
                readPGMHeader(rd);
                pixels = new int[width][height];
                in.skipBytes(header.length);
                fillPGMPixels(in);
                break;
        }
        
        in.close();
    }
    
    /**
     * Create an exact duplicate of an image - all header fields are copied, with the option of copying pixel values
     */
    public Image(Image im, boolean copyPixels) {
        height = im.height;
        width = im.width;
        bitDepth = im.bitDepth;
        type = im.type;
        filesize = im.filesize;
        header = im.header;
        colorTable = im.colorTable;
        pixels = new int[width][height];
        if(copyPixels) {
            for(int j=0; j < height; j++)
                for(int i=0; i < width; i++)
                    pixels[i][j] = im.pixels[i][j];
        }
    }
    
    /**
     * Writes the image to a file
     */
    public void save(String filename) throws IOException {
        DataOutputStream out = new DataOutputStream(new FileOutputStream(filename));
        
        // first write header to file
        out.write(header);
        
        switch(type) {
            case BITMAP: saveBitmapPixels(out); break;
            case PGM: savePGMPixels(out); break;
        }
        
        out.close();
    }
    
    private void saveBitmapPixels(DataOutputStream out) throws IOException {
        // The bitmap file format requires that each line of pixels be a multiple of 4 bytes, and if it
        // is not, then 0's are added until a multiple of 4 bytes is reached. The zeroPaddedBytes variable
        // keeps track of how many of these 0's are included, based on the width dimension of the bitmap,
        // and will skip over these bytes when reading from the file because they do not represent pixel data.
        int zeroPaddedBytes = (width * (bitDepth/8)) % 4 == 0 ? 0 : (4 - ((width * (bitDepth/8)) % 4));
        byte[] zeroes = new byte[zeroPaddedBytes];
        
        for(int j=height-1; j >= 0; j--) { // iterate through each row (reverse ordering)
            for(int i=0; i < width; i++) { // iterate through each pixel in this row
                switch(bitDepth) {
                    case 8: out.writeByte((byte)pixels[i][j]); break;
                    case 16: out.writeShort((short)pixels[i][j]); break;
                    case 24: out.writeShort(pixels[i][j] >> 8 & 0x0000FFFF); out.writeByte(pixels[i][j] & 0x000000FF); break;
                    default: throw new IOException("Unsupported bit-depth: "+bitDepth+". Only 8-, 16-, and 24-bit bitmaps are allowed.");
                }
            }
            // write all zero-padded bytes in this line
            out.write(zeroes);
        }
    }

    private void savePGMPixels(DataOutputStream out) throws IOException {
        for(int j=0; j < height; j++) { // iterate through each row (top to bottom)
            for(int i=0; i < width; i++) { // iterate through each pixel in this row (left to right)
                out.writeByte((byte)pixels[i][j]);
            }
        }
    }

    /**
     * Reads in byte-by-byte the bitmap header data. Also extracts width, height, and
     * bit-depth information of the bitmap from the header.
     */
    private void readBitmapHeader(DataInputStream in) throws IOException {
        try {
            int offset = 0; // counter for current byte # in header
            
            // two placeholder for temporarily storing up to two and four bytes of data
            byte[] buffer2 = new byte[2];
            byte[] buffer4 = new byte[4];
            
            // first 2 bytes = file format code; must be 'BM' for bitmap
            byte[] bfType = new byte[2];
            in.readFully(bfType);
            if(!(bfType[0] == 'B' && bfType[1] == 'M')) { // Make sure first 2 bytes = 'BM', indicating bitmap format
                in.close();
                throw new IOException("Expected bitmap format, but header is not recognized.");
            }
            
            // next 4 bytes = file size in bytes
            byte[] bfSize = new byte[4];
            in.readFully(bfSize);
            filesize = (bfSize[3] << 24 & 0xFF000000) | (bfSize[2] << 16 & 0x00FF0000) | (bfSize[1] << 8 & 0x0000FF00) | (bfSize[0] & 0x000000FF);
            
            // skip next 4 bytes
            byte[] bfReserved = new byte[4];
            in.readFully(bfReserved);
            
            // next 4 bytes = byte offset from beginning of file at which the pixel data begins (the header size in bytes)
            in.readFully(buffer4);
            int headerSize = (buffer4[3] << 24 & 0xFF000000) | (buffer4[2] << 16 & 0x00FF0000) | (buffer4[1] << 8 & 0x0000FF00) | (buffer4[0] & 0x000000FF);
            
            // now write all previous bytes into the header[] array
            header = new byte[headerSize];
            Image.append(header, bfType, offset); offset+=2;
            Image.append(header, bfSize, offset); offset+=4;
            Image.append(header, bfReserved, offset); offset+=4;
            Image.append(header, buffer4, offset); offset+=4;
            
            // skip next 4 bytes
            in.readFully(buffer4);
            append(header, buffer4, offset); offset+=4;
            
            // next 4 bytes = width of bitmap in pixels
            in.readFully(buffer4);
            width = (buffer4[3] << 24 & 0xFF000000) | (buffer4[2] << 16 & 0x00FF0000) | (buffer4[1] << 8 & 0x0000FF00) | (buffer4[0] & 0x000000FF);
            append(header, buffer4, offset); offset+=4;
            
            // next 4 bytes = height of bitmap in pixels
            in.readFully(buffer4);
            height = (buffer4[3] << 24 & 0xFF000000) | (buffer4[2] << 16 & 0x00FF0000) | (buffer4[1] << 8 & 0x0000FF00) | (buffer4[0] & 0x000000FF);
            append(header, buffer4, offset); offset+=4;
            
            // skip next 2 bytes
            in.readFully(buffer2);
            append(header, buffer2, offset); offset+=2;
            
            // next 2 bytes = bit-depth of the bitmap
            in.readFully(buffer2);
            bitDepth = (short)((buffer2[1] << 8 & 0x0000FF00) | (buffer2[0] & 0x000000FF));
            if(bitDepth < 8 || bitDepth > 24) throw new IOException("Unsupported bit-depth: "+bitDepth+". Only 8-, 16-, and 24-bit bitmaps are allowed.");
            append(header, buffer2, offset); offset+=2;
            
            // next 4 bytes = 0 if no RLE compression
            in.readFully(buffer4);
            if(!(buffer4[0] == 0 && buffer4[1] == 0 && buffer4[2] == 0 && buffer4[3] == 0)) {
                in.close();
                throw new IOException("This bitmap is RLE compressed. Only uncompressed bitmaps are allowed.");
            }
            append(header, buffer4, offset); offset+=4;
            
            // skip next 20 bytes
            for(int i=0; i < 20; i += 4) {
                in.readFully(buffer4);
                append(header, buffer4, offset); offset+=4;
            }
            
            // now read in the color table data (if not 24-bit)
            if(bitDepth < 24) {
                colorTable = new int[(int)Math.pow(2, bitDepth)];
                
                for(int i=0; offset < headerSize; i++) {
                    int b = in.readUnsignedByte(); header[offset] = (byte)b; offset++;
                    int g = in.readUnsignedByte(); header[offset] = (byte)g; offset++;
                    int r = in.readUnsignedByte(); header[offset] = (byte)r; offset++;
                    byte temp = in.readByte(); header[offset] = temp; offset++;
                    
                    int rgb = (r << 16) | (g << 8) | b;
                    colorTable[i] = rgb;
                }
            }

            assert (offset == headerSize);
        
        } catch(EOFException e) {
            // if an EOFException is thrown, the header is not properly formatted
            throw new IOException("Invalid bitmap file. Header is not recognized.");
        }
    }
    
    /**
     * Reads in byte-by-byte the PGM header data. Also extracts width and height information of 
     * the PGM from the header.
     */
    private void readPGMHeader(BufferedReader rd) throws IOException {
        bitDepth = 8;
        
        try {
            
            ArrayList items = new ArrayList(); // for temp storage of header items
            
            // temporarily store header as one long ASCII string
            String s = "";
            
            // placeholder for temporarily storing one line
            String currLine = rd.readLine();
            s += currLine+"\n";
            
            // skip any comment lines
            while(currLine.startsWith("#")) {
                currLine = rd.readLine();
                s += currLine+"\n";
            }
            
            // split up next line based on white-space
            StringTokenizer tkn = new StringTokenizer(currLine);
            
            // first 2 bytes = file format code; must be 'P5' for PGM
            if(!tkn.nextToken().equals("P5")) {
                rd.close();
                throw new IOException("Expected PGM format, but header is not recognized.");
            }
            while(tkn.hasMoreTokens()) {
                items.add(tkn.nextToken());
            }
            
            // skip any comment lines
            while((currLine = rd.readLine()).startsWith("#")) {
                s += currLine+"\n";
            }
            
            // get rest of header
            while(items.size() < 3) {
                s += currLine+"\n";
                tkn = new StringTokenizer(currLine);
                while(tkn.hasMoreTokens()) {
                    items.add(tkn.nextToken());
                }
                    
                // skip any comment lines
                while((currLine = rd.readLine()).startsWith("#")) {
                    s += currLine+"\n";
                }
            }
                
            // copy over header items
            // width should be first
            assert ((String)items.get(0)).matches("\\d+");
            width = Integer.parseInt((String)items.get(0));
            // height should be next
            assert ((String)items.get(1)).matches("\\d+");
            height = Integer.parseInt((String)items.get(1));
                
            // copy over header string to header byte array
            header = new byte[s.length()];
            for(int i=0; i < header.length; i++)
                header[i] = (byte)s.charAt(i);
            
        } catch(EOFException e) {
            // if an EOFException is thrown, the header is not properly formatted
            throw new IOException("Invalid PGM file. Header is not recognized.");
        }
    }
    
    private void fillBitmapPixels(DataInputStream in) throws IOException {
        
        // The bitmap file format requires that each line of pixels be a multiple of 4 bytes, and if it
        // is not, then 0's are added until a multiple of 4 bytes is reached. The zeroPaddedBytes variable
        // keeps track of how many of these 0's are included, based on the width dimension of the bitmap,
        // and will skip over these bytes when reading from the file because they do not represent pixel data.
        int zeroPaddedBytes = (width * (bitDepth/8)) % 4 == 0 ? 0 : (4 - ((width * (bitDepth/8)) % 4));
        
        for(int j=height-1; j >= 0; j--) { // iterate through each row (reverse ordering)
            for(int i=0; i < width; i++) { // iterate through each pixel in this row
                int x;
                switch(bitDepth) {
                    case 8: x = in.readUnsignedByte(); break;
                    case 16: x = in.readUnsignedShort(); break;
                    case 24: x = (in.readUnsignedShort() << 8) | in.readUnsignedByte(); break;
                    default: throw new IOException("Unsupported bit-depth: "+bitDepth+". Only 8-, 16-, and 24-bit bitmaps are allowed.");
                }
                pixels[i][j] = x;
            }
            // skip over all zero-padded bytes in this line
            in.skipBytes(zeroPaddedBytes);
        }
    }
    
    private void fillPGMPixels(DataInputStream in) throws IOException {
        
        for(int j=0; j < height; j++) { // iterate through each row (top to bottom)
            for(int i=0; i < width; i++) { // iterate through each pixel in this row (left to right)
                pixels[i][j] = in.readUnsignedByte();
            }
        }
    }
    
    public int getRGB(int pixel) {
        int rgb = pixel;
        
        switch(type) {
            case BITMAP:
                if(bitDepth != 24) rgb = colorTable[pixel]; break;
        }
        
        return rgb;
    }
    
    private void getType(String filename) throws IOException {
        // determine image type from file suffix
        int index = filename.lastIndexOf('.');
        if(index == -1)
            throw new IOException("Image does not have a file extension.");
        
        String suffix = filename.substring(index+1).toLowerCase();
        if(suffix.equals("bmp") || suffix.equals("dib") || suffix.equals("rle")) type = BITMAP;
        else if(suffix.equals("pgm")) type = PGM;
        else throw new IOException("Unrecognized file extension.");
    }

    /**
     * Appends a byte array onto the end of another byte array.
     * 
     * @param b the byte array to add to
     * @param c the array which is being appended
     * @param offset the index at which to begin appending
     */
    private static void append(byte[] b, byte[] c, int offset) {
        for(int i=0; i < c.length; i++)
            b[offset+i] = c[i];
    }
}